/*
 * FlyHow filter for AviSynth
 * Arne Ansper - arne@cyber.ee
 *
 * Based on AudioGraph filter for AviSynth
 * Richard Ling - r.ling(at)eudoramail.com
 * 
 * USAGE:
 * ------
 *
 *   LoadPlugin("flyhow.dll")
 *   audio = WAVSource("sample.wav")
 *   return FlyHow(AudioDub(BlankClip(1000), audio), 20)
 * 
 * TO DO:
 * -----
 */

#include "windows.h"
#include "crtdbg.h"
#include "avisynth.h"
#include "audio.h"
#include "kiss_fft.h"
#include "kiss_fftr.h"

/**
  * Helper class to anti-alias text
 **/
class Antialiaser 
{  
public:
	Antialiaser(int width, int height, const char fontname[], int size);
	virtual ~Antialiaser();
	HDC GetDC();

	void Apply(const VideoInfo& vi, PVideoFrame* frame, int pitch, int textcolor, int halocolor);
	void ApplyYV12(BYTE* buf, int pitch, int textcolor, int halocolor, int UVpitch,BYTE* bufV,BYTE* bufU);
	void ApplyYUY2(BYTE* buf, int pitch, int textcolor, int halocolor);
	void ApplyRGB24(BYTE* buf, int pitch, int textcolor, int halocolor);
	void ApplyRGB32(BYTE* buf, int pitch, int textcolor, int halocolor);  

private:
	const int w, h;
	HDC hdcAntialias;
	HBITMAP hbmAntialias;
	void* lpAntialiasBits;
	HFONT hfontDefault;
	HBITMAP hbmDefault;
	char* alpha_bits;
	bool dirty;

	void GetAlphaRect();
};

int rudder = 0;
int left_aileron = 1;
int left_flap = 2;
int elevator = 3;
int right_flap = 4;
int right_aileron = 5;
int invalid = 6;

struct channel_params {
	channel_params() {channel = 0;}

	int function;
	int channel;
	int cup;
	int cdown;
	int s0;
	int sup;
	int sdown;

	int left_x;
	int right_x;
	int y0;

	int down_mul;
	int down_div;
	int down_mul_ms;
	int up_mul;
	int up_div;
	int up_mul_ms;

	int top_y;
	int bottom_y;
	int x0;
};

class FlyHow : public GenericVideoFilter
{
public:
	FlyHow(PClip _child, int size, int pfreq, int ppoints, IScriptEnvironment* _env);
	virtual ~FlyHow();
	void GetChannels(int frame, IScriptEnvironment* env);
	int GetPeak(int frame, IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrameCalibrate(int n, IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrameDoppler(int n, IScriptEnvironment* env);

	channel_params* cp;
	int freq;
	int cached_peak;
	int fft_points;

private:
	int getsample(int sample);
	void get_data(int n);
	void setup_geometry(int width, int height,IScriptEnvironment* _env);

	/* what channel (L or R) to use */
	int input_channel;
	BYTE* audio_buffer;
	int cache_lookup;
	
	int samples_per_frame;
	int bytes_per_sample;
	int frames_back;

	/* number of channels in use */
	int num_channels;
	/* values for channels in microseconds (0 - center) */
	int channel_values[12];
	/* center values for channels */
	int channel_centers[12];
	int channel_mins[12];
	int channel_maxs[12];

	Antialiaser antialiaser;

	int max_sample;
	int min_sample;
	int mid_sample;
	int start_sample;
	int geometry_done;

	int show_values;
};

/*
 * FlyHow::FlyHow
 * 
 * Parameters:
 *   _child				 The PClip which whose audio is to be graphed.
 */
FlyHow::FlyHow(PClip _child, int _size, int pfreq, int ppoints, IScriptEnvironment* _env) :
	GenericVideoFilter(ConvertAudio::Create(_child,SAMPLE_INT16|SAMPLE_INT8,SAMPLE_INT16)),
	antialiaser(vi.width, vi.height, "Arial", _size),
	audio_buffer(NULL),
	cache_lookup(-1),
	input_channel(0),
	num_channels(8),
	cp(NULL),
	geometry_done(0),
	show_values(_size != 0),
	freq(pfreq),
	cached_peak(-1),
	fft_points(1 << ppoints)
{
	if (!vi.IsRGB32()) 
		_env->ThrowError("FlyHow: Only RGB32 video is supported at the moment.");

	if (!vi.HasAudio())
		_env->ThrowError("FlyHow: clip has no audio");

	/*
	 * Allocate the buffer for raw audio data.  We only ever read raw audio
	 * data for one frame at a time.
	 */
	bytes_per_sample = vi.BytesPerAudioSample();

	if (freq == -1) {
		frames_back = 1;
		samples_per_frame = (int)vi.AudioSamplesFromFrames(1) * (frames_back + 1);
	} else {
		/* how many frames we must go back to get enough data for FFT */
		frames_back = fft_points / (int)vi.AudioSamplesFromFrames(1);
		samples_per_frame = (int)vi.AudioSamplesFromFrames(1) * (frames_back + 1);
		frames_back /= 2; /* so that we are in the middle. might cause crash */
	}
	audio_buffer = new BYTE[bytes_per_sample * samples_per_frame];

	for (int i = 0; i < sizeof(channel_centers) / sizeof(int); i++) {
		channel_centers[i] = 1520;
		channel_mins[i] = MAX_INT;
		channel_maxs[i] = MIN_INT;
	}
}

/* this method sets up the positions of control surfaces */
void FlyHow::setup_geometry(int width, int height,IScriptEnvironment* env)
{
	int max_down = MIN_INT;	/* max_down is biggest number in mm */
	int max_up = MAX_INT; /* max_up is smallest number in mm */

	int cf;

	/* dimension the horizontal surfaces. find max up and down deflections. 
	 * NB! the rule is that max_down > max_up  
	 * XXX we should throw error if it is not so. 
	 * NB! start from left_aileron, omit the rudder! */
	for (cf = left_aileron; cf < invalid; cf+= 1) {
		if (cp[cf].channel != 0) {
			if (cp[cf].cdown > max_down) {
				max_down = cp[cf].cdown;
			}
			if (cp[cf].cup < max_up) {
				max_up = cp[cf].cup;
			}
		}
	}

	int h_spacer = width / 20;
	int h_control = (width * (20 - 6)) / (5 * 20);

	/* vertical spacing for horizontal controls - we have 1/20 frame around controls
	 * remaining height is divided between 2/3 horizontal and 1/3 vertical controls and there
	 * is 1/20 between them. the neutral line is drawn accoring to max_up and max_down
	 */
	int v_horizontal = (height * 17 * max_down * 2) / (20 * 3 * (max_down - max_up));

	for (cf = left_aileron; cf < invalid; cf++) {
		if (cp[cf].channel != 0) {
			cp[cf].left_x = cf * h_spacer + (cf - 1) * h_control;
			cp[cf].right_x = cf * h_spacer + cf * h_control;
			cp[cf].y0 = v_horizontal + height / 20;

			/* set up scaling factors for both up and down movements */
			cp[cf].down_mul = cp[cf].cdown * v_horizontal;
			cp[cf].down_div = max_down * (cp[cf].sdown - cp[cf].s0);
			cp[cf].down_mul_ms = cp[cf].cdown * max_down;
			cp[cf].up_mul = cp[cf].cup * ((height * 17 * 2) / (20 * 3) - v_horizontal);
			cp[cf].up_div = max_up * (cp[cf].sup - cp[cf].s0);
			cp[cf].up_mul_ms = cp[cf].cup * max_up;
		}
	}

	/* rudder is drawn vertically. there are no other vertical control surfaces, so the 
	 * math is simpler. we make the rudder half the width of the elevator */
	if (cp[rudder].channel != 0) {
		cp[rudder].top_y = (height * 19) / 20;
		cp[rudder].bottom_y = (height * 17 * 2) / (20 * 3) + (2 * height) / 20;
		cp[rudder].x0 = 3 * h_spacer + (5 * h_control) / 2;

		/* down is right */
		cp[rudder].down_mul = (h_control * cp[rudder].cdown) / (2 * (cp[rudder].cdown - cp[rudder].cup));
		cp[rudder].down_div = (cp[rudder].sdown - cp[rudder].s0);
		cp[rudder].down_mul_ms = cp[rudder].cdown;
		cp[rudder].up_mul = (h_control * cp[rudder].cup) / (2 * (cp[rudder].cdown - cp[rudder].cup));
		cp[rudder].up_div = (cp[rudder].sup - cp[rudder].s0);
		cp[rudder].up_mul_ms = cp[rudder].cup;
	}

	geometry_done = 1;
}


/*
 * FlyHow::~FlyHow
 * 
 * Free allocated memory.
 */
FlyHow::~FlyHow()
{
	delete[] audio_buffer;
	delete cp;
}

int FlyHow::GetPeak(int frame, IScriptEnvironment* env)
{
	if (frame < frames_back) {
		return freq;
	}

	if (cache_lookup != frame)
	{
		child->GetAudio(audio_buffer, vi.AudioSamplesFromFrames(frame - frames_back), samples_per_frame, env);
		
		/* Perform FFT */
		kiss_fftr_cfg cfg = kiss_fftr_alloc(fft_points, 0, 0, 0);
		kiss_fft_scalar *tbuf = new kiss_fft_scalar[fft_points];
		kiss_fft_cpx *fbuf = new kiss_fft_cpx[fft_points/2 + 1];

		for (int i = 0;i < fft_points;i++) {
			tbuf[i] = getsample(i);
		}
		kiss_fftr(cfg, tbuf, fbuf);

		int peak = 0;
		double peak_value = 0;

		/* Find peak */
		for (i = 4 * freq / 5;i < 6 * freq / 5; i++) {
			double np = fbuf[i].r * fbuf[i].r + fbuf[i].i * fbuf[i].i;
			if (np > peak_value) {
				peak_value = np;
				peak = i;
			}
		}
		free(cfg);
		free(tbuf);
		free(fbuf);

		cache_lookup = frame;
		cached_peak = peak;
	}
	return cached_peak;
}

/*
 * FlyHow::GetChannels
 * 
 * Fill channel_values array from the audioframe corresponding to the given video frame. If
 * the data already exists in the cache, it is returned immediately.
 * Otherwise it is generated from the raw audio data, stored in the cache, and
 * returned.
 *
 * Parameters:
 *   frame	  The frame whose audioframe is needed.
 *   env		A pointer to the IScriptEnvironment.
 * 
 */
void FlyHow::GetChannels(int frame, IScriptEnvironment* env)
{
	if (cache_lookup != frame) {
		child->GetAudio(audio_buffer, vi.AudioSamplesFromFrames(frame - frames_back), samples_per_frame, env);
		/* extract channel information */
		get_data(frame);
		cache_lookup = frame;
	}
}

PVideoFrame __stdcall FlyHow::GetFrame(int n, IScriptEnvironment* env)
{
	if (freq != -1) {
		return GetFrameDoppler(n, env);
	}
	if (cp == NULL) {
		return GetFrameCalibrate(n, env);
	}

	/* animated output. SimpleSample used as example. comments removed */
	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);
	
	const unsigned char* srcp = src->GetReadPtr();
	unsigned char* dstp = dst->GetWritePtr();
	
	const int dst_pitch = dst->GetPitch();
	const int dst_width = dst->GetRowSize();
	const int dst_height = dst->GetHeight();

	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	
	if (n < frames_back) {
		return dst;
	}

	GetChannels(n, env);

	if (!geometry_done) {
		setup_geometry(dst_width / 4, dst_height,env);
	}

	int w, h;
	HDC hdc;
	char text[160];

	if (show_values) {
		hdc = antialiaser.GetDC();
		SetTextAlign(hdc, TA_BASELINE | TA_CENTER);
		RECT r = { 0, 0, 32767, 32767 };
		FillRect(hdc, &r, (HBRUSH)GetStockObject(BLACK_BRUSH));
	}

	if (vi.IsRGB32()) {
		for (h=0; h < src_height;h++) {
			for (w = 0; w < src_width/4; w+=1) {
				*((unsigned int *)dstp + w) = *((unsigned int *)srcp + w);
			}
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		}
		
		// now draw horizontal controls
		for (int cf = left_flap; cf < invalid; cf++) {
			if (cp[cf].channel != 0) {
				int startline;
				int endline;
				// channel delta
				int cvd = channel_values[cp[cf].channel - 1] - cp[cf].s0;
				int ms;
				unsigned int color;

				/* up or down? */
				if (cp[cf].sdown > cp[cf].sup ? (cvd > 0) : (cvd < 0)) {
					/* down */
					startline = cp[cf].y0 - (cvd * cp[cf].down_mul) / cp[cf].down_div;
					endline = cp[cf].y0;
					ms = (cvd * cp[cf].down_mul_ms) / cp[cf].down_div;
					color = 0x0000FF00;
				} else {
					/* up */
					endline = cp[cf].y0 + (cvd * cp[cf].up_mul) / cp[cf].up_div;
					startline = cp[cf].y0;
					ms = (cvd * cp[cf].up_mul_ms) / cp[cf].up_div;
					color = 0x00FF0000;
				}

				if (endline == startline) {
					endline ++;
					color = 0x0000FF00;
				}
				dstp = dst->GetWritePtr();
				dstp = dstp + startline * dst_pitch;

				for (h = 0; h < endline - startline; h++) {
					for (w = cp[cf].left_x; w < cp[cf].right_x; w++) { 
						*((unsigned int *)dstp + w) = color;
					}	
					dstp = dstp + dst_pitch; 
				}
				if (show_values) {
					wsprintf(text, "%d", ms);
					TextOut(hdc, (8 * (cp[cf].left_x + cp[cf].right_x + 1)) / 2, 
							8 * (dst_height - cp[cf].y0 - 1),
							text, strlen(text));
				}
			}
		}

		// now draw rudder
		if (cp[rudder].channel != 0) {
			int startx;
			int endx;
			int ms;

			// channel delta
			int cvd = channel_values[cp[rudder].channel - 1] - cp[rudder].s0;
			unsigned int color;

			/* up or down? */
			if (cp[rudder].sdown > cp[rudder].sup ? (cvd > 0) : (cvd < 0)) {
				/* down */
				endx = cp[rudder].x0 + (cvd * cp[rudder].down_mul) / cp[rudder].down_div;
				startx = cp[rudder].x0;
				ms = (cvd * cp[rudder].down_mul_ms) / cp[rudder].down_div;
				color = 0x0000FF00;
			} else {
				/* up */
				startx = cp[rudder].x0 + (cvd * cp[rudder].up_mul) / cp[rudder].up_div;
				endx = cp[rudder].x0;
				ms = (cvd * cp[rudder].up_mul_ms) / cp[rudder].up_div;
				color = 0x00FF0000;
			}

			if (startx == endx) {
				endx++;
				color = 0x0000FF00;
			}
			dstp = dst->GetWritePtr();
			dstp = dstp + cp[rudder].bottom_y * dst_pitch;

			for (h = 0; h < cp[rudder].top_y - cp[rudder].bottom_y; h++) {
				for (w = startx; w < endx; w++) { 
					*((unsigned int *)dstp + w) = color;
				}	
				dstp = dstp + dst_pitch; 
			}
			if (show_values) {
				wsprintf(text, "%d", ms);
				TextOut(hdc, 8 * (cp[rudder].x0 + 3), 8 * dst_height - 
						(8 * (cp[rudder].top_y + 5 * cp[rudder].bottom_y) / 6), 
						text, strlen(text));
			}
		}
		if (show_values) {
			GdiFlush();
			antialiaser.Apply(vi, &dst, dst->GetPitch(),
					vi.IsYUV() ? 0xD21092 : 0xFFFF00, vi.IsYUV() ? 0x108080 : 0);
		}
	}

	return dst;
}


PVideoFrame __stdcall FlyHow::GetFrameCalibrate(int n, IScriptEnvironment* env)
{
	PVideoFrame frame = child->GetFrame(n, env);
	env->MakeWritable(&frame);

	if (n < frames_back) {
		return frame;
	}

	GetChannels(n, env);
	HDC hdc = antialiaser.GetDC();
	SetTextAlign(hdc, TA_BASELINE | TA_LEFT);
	RECT r = { 0, 0, 32767, 32767 };
	FillRect(hdc, &r, (HBRUSH)GetStockObject(BLACK_BRUSH));

	char text[160];
	int line = 1;
#if 0
	wsprintf(text, "%d", n);
	TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 1024, 128 * (line++), text, strlen(text));
	wsprintf(text, "%d", min_sample);
	TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 1024, 128 * (line++), text, strlen(text));
	wsprintf(text, "%d", max_sample);
	TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 1024, 128 * (line++), text, strlen(text));
	wsprintf(text, "%d", mid_sample);
	TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 1024, 128 * (line++), text, strlen(text));
	wsprintf(text, "%d", start_sample);
	TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 1024, 128 * (line++), text, strlen(text));
#endif

	for (int i = 0;i < num_channels; i++) {
		wsprintf(text, "Ch%d: %8d %8d %8d", i + 1, 
				channel_values[i],
				channel_mins[i],
				channel_maxs[i]);
		TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 2048, 128 * (line++), text, strlen(text));
	}
	GdiFlush();

	antialiaser.Apply(vi, &frame, frame->GetPitch(),
			vi.IsYUV() ? 0xD21092 : 0xFFFF00, vi.IsYUV() ? 0x108080 : 0);

	return frame;
}

PVideoFrame __stdcall FlyHow::GetFrameDoppler(int n, IScriptEnvironment* env)
{
	PVideoFrame frame = child->GetFrame(n, env);
	env->MakeWritable(&frame);

	int peak = GetPeak(n, env);

	HDC hdc = antialiaser.GetDC();
	SetTextAlign(hdc, TA_BASELINE | TA_LEFT);
	RECT r = { 0, 0, 32767, 32767 };
	FillRect(hdc, &r, (HBRUSH)GetStockObject(BLACK_BRUSH));

	char text[160];

	wsprintf(text, "Peak: %d. Speed: %d m/s", peak,
			(int)(337.0 * (1 - ((double)freq)/peak)));
	TextOut(hdc, child->GetParity(n) ? 32 : vi.width*8 - 2048, 128, text, strlen(text));
	GdiFlush();

	antialiaser.Apply(vi, &frame, frame->GetPitch(),
			vi.IsYUV() ? 0xD21092 : 0xFFFF00, vi.IsYUV() ? 0x108080 : 0);

	return frame;
}

/*
 * This plugin creates full-frame picture of the gliders control surfaces. You can position
 * it using the Overlay and also get nice semi-transparent effects.
 *
 * You must first create empty video clip with suitable dimensions. Up and left deflections will
 * be drawn in pure red, down and right in pure green.
 */



/*
 * Create_FlyHow
 * 
 * Called by AviSynth to create a new FlyHow filter.
 * 
 * Parameters:
 *   args		   The arguments to the filter, specified in the script.
 *   user_data	  Our user data parameter; not used.
 *   env			A pointer to the IScriptEnvironment.
 * 
 * Returns:
 *   An AVSValue__cdecl which does something.
 */
AVSValue __cdecl Create_FlyHowCalibrate(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new FlyHow(args[0].AsClip(), 128, -1, 10, env);
}

AVSValue __cdecl Create_Doppler(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	FlyHow* fh = new FlyHow(args[0].AsClip(), args[1].AsInt(), args[2].AsInt(), args[3].AsInt(), env);

	return fh;
}

int decodename(const char* name)
{
	int retval = invalid;

	if (strcmp(name, "rudder") == 0) {
		retval = rudder;
	}
	if (strcmp(name, "elevator") == 0) {
		retval = elevator;
	}
	if (strcmp(name, "left_aileron") == 0) {
		retval = left_aileron;
	}
	if (strcmp(name, "right_aileron") == 0) {
		retval = right_aileron;
	}
	if (strcmp(name, "right_flap") == 0) {
		retval = right_flap;
	}
	if (strcmp(name, "left_flap") == 0) {
		retval = left_flap;
	}
	return(retval);
}


AVSValue __cdecl Create_FlyHow(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	channel_params* cp = new channel_params[invalid + 1];
	FlyHow* fh = new FlyHow(args[0].AsClip(), args[1].AsInt(), -1, 10, env);

	for (int i = 0;i < 6; i++) {
		int func = decodename(args[i * 7 + 2].AsString());
		// XXX more parameter errors reporting is needed here
		if (func < invalid) {
			cp[func].function = func;
			int channel = args[i * 7 + 3].AsInt();
			if (channel < 0 || channel > 10) {
				env->ThrowError("FlyHow: Invalid channel number");
			}
			cp[func].channel = channel;
			cp[func].s0 = args[i * 7 + 4].AsInt();
			int cup = args[i * 7 + 5].AsInt();
			if (cup >= 0) {
				env->ThrowError("FlyHow: Cup must be negative");
			}
			cp[func].cup = cup;
			int cdown = args[i * 7 + 6].AsInt();
			if (cdown <= 0) {
				env->ThrowError("FlyHow: Cdown must be positive");
			}
			cp[func].cdown = cdown;
			cp[func].sup = args[i * 7 + 7].AsInt();
			cp[func].sdown = args[i * 7 + 8].AsInt();
		}
	}
	fh->cp = cp;

	return fh;
}


/*
 * AvisynthPluginInit
 * 
 * Called by AviSynth when the plugin is loaded.  It registers a new function
 * called "FlyHow", which scripts can then call to invoke this filter.
 *
 * Parameters:
 *   env	A pointer to the IScriptEnvironment.
 * 
 * Returns:
 *   A short description of the plugin.
 */
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	env->AddFunction("FlyHowCalibrate", "c", Create_FlyHowCalibrate, NULL);
	env->AddFunction("FlyHow", "ci"
			"siiiiii"
			"siiiiii"
			"siiiiii"
			"siiiiii"
			"siiiiii"
			"siiiiii"
			, Create_FlyHow, NULL);
	env->AddFunction("Doppler", "ciii", Create_Doppler, NULL);
	return "`FlyHow' sample plugin";
}

/* get sample of the specified channel.
 *
 * uses audio_buffer and input_channel */
int FlyHow::getsample(int sample)
{
	int retval;

	BYTE* src = audio_buffer + sample * bytes_per_sample;
	if (vi.SampleType() == SAMPLE_INT16) {
		src += input_channel * 2;
		retval = *(short*)src;
	} else {
		src += input_channel;
		retval = *(BYTE*)src - 128;
	}
	return(retval);
}

/* computes the values for channels from the audio data. 
 * all calculations are done in nanoseconds - this allows us to use 
 * 32 bit integer calculations without loosing much precision.
 *
 * uses - audio_buffer
 * sets - num_channels, channel_values */
void FlyHow::get_data(int frame)
{
	int prev_sample = MAX_INT;
	int i;
	int channel = -1;
	int nanoseconds_per_sample = 1000000000L / vi.SamplesPerSecond();
	int last_front_time = 0;
	int current_time = 0;

	max_sample = MIN_INT;
	min_sample = MAX_INT;

	/* find peaks and middle value*/
	for(i = 0; i < samples_per_frame ; i++) {
		int sample = getsample(i);

		if (sample > max_sample) max_sample = sample;
		if (sample < min_sample) min_sample = sample;
	}
	mid_sample = (max_sample + min_sample) / 2;

	/* decode channels, look for first full frame */
	for(i = 0; i < samples_per_frame ; i++) {
		int sample = getsample(i);

		/* if pause is more than 4ms then we are in 
		 * the start of the new frame */
		if (current_time - last_front_time > 3000000){
			if (channel < 0) {
				channel = 0; // start
				start_sample = i;
			}
			if (channel > 0) {
				break; // end
			}
		}
		if (sample >= mid_sample && prev_sample < mid_sample) {
			int crosspoint = (sample - mid_sample) * nanoseconds_per_sample / 
					(sample - prev_sample);

			if (channel >= 0){
				/* skip the first front */
				if(channel > 0) {
					channel_values[channel - 1] = (current_time - 
							crosspoint - last_front_time) / 1000 - 
							channel_centers[channel - 1];
					if (channel_values[channel - 1] > channel_maxs[channel - 1]) { 
						channel_maxs[channel - 1] = channel_values[channel - 1];
					}
					if (channel_values[channel - 1] < channel_mins[channel - 1]) { 
						channel_mins[channel - 1] = channel_values[channel - 1];
					}
				}
				channel++;
				/* too much channels */
				if (channel > 10) { 
					break;
				}
			}
			last_front_time = current_time - crosspoint;
		}
		prev_sample = sample;
		current_time += nanoseconds_per_sample;
	}
	num_channels = channel - 1;
}

inline static HFONT LoadFont(const char name[], int size, bool bold, bool italic) 
{
  return CreateFont( size, 0, 0, 0, bold ? FW_BOLD : FW_NORMAL,
                     italic, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
                     CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FF_DONTCARE | DEFAULT_PITCH, name );
}

/******************************
 *******   Anti-alias    ******
 *****************************/
Antialiaser::Antialiaser(int width, int height, const char fontname[], int size)
 : w(width), h(height)
{
  struct {
    BITMAPINFOHEADER bih;
    RGBQUAD clr[2];
  } b;

  b.bih.biSize                    = sizeof(BITMAPINFOHEADER);
  b.bih.biWidth                   = width * 8 + 32;
  b.bih.biHeight                  = height * 8 + 32;
  b.bih.biBitCount                = 1;
  b.bih.biPlanes                  = 1;
  b.bih.biCompression             = BI_RGB;
  b.bih.biXPelsPerMeter   = 0;
  b.bih.biYPelsPerMeter   = 0;
  b.bih.biClrUsed                 = 2;
  b.bih.biClrImportant    = 2;
  b.clr[0].rgbBlue = b.clr[0].rgbGreen = b.clr[0].rgbRed = 0;
  b.clr[1].rgbBlue = b.clr[1].rgbGreen = b.clr[1].rgbRed = 255;

  hdcAntialias = CreateCompatibleDC(NULL);
  hbmAntialias = CreateDIBSection
    ( hdcAntialias,
      (BITMAPINFO *)&b,
      DIB_RGB_COLORS,
      &lpAntialiasBits,
      NULL,
      0 );
  hbmDefault = (HBITMAP)SelectObject(hdcAntialias, hbmAntialias);

  HFONT newfont = LoadFont(fontname, size, true, false);
  hfontDefault = (HFONT)SelectObject(hdcAntialias, newfont);

  SetMapMode(hdcAntialias, MM_TEXT);
  SetTextColor(hdcAntialias, 0xffffff);
  SetBkColor(hdcAntialias, 0);

  alpha_bits = new char[width*height*2];

  dirty = true;
}


Antialiaser::~Antialiaser() {
  DeleteObject(SelectObject(hdcAntialias, hbmDefault));
  DeleteObject(SelectObject(hdcAntialias, hfontDefault));
  DeleteDC(hdcAntialias);
  delete[] alpha_bits;
}


HDC Antialiaser::GetDC() {
  dirty = true;
  return hdcAntialias;
}


void Antialiaser::Apply( const VideoInfo& vi, PVideoFrame* frame, int pitch, int textcolor, 
                         int halocolor ) 
{
  if (vi.IsRGB32())
    ApplyRGB32((*frame)->GetWritePtr(), pitch, textcolor, halocolor);
  else if (vi.IsRGB24())
    ApplyRGB24((*frame)->GetWritePtr(), pitch, textcolor, halocolor);
  else if (vi.IsYUY2())
    ApplyYUY2((*frame)->GetWritePtr(), pitch, textcolor, halocolor);
  else if (vi.IsYV12())
    ApplyYV12((*frame)->GetWritePtr(), pitch, textcolor, halocolor, (*frame)->GetPitch(PLANAR_U),(*frame)->GetWritePtr(PLANAR_U),(*frame)->GetWritePtr(PLANAR_V));
}

void Antialiaser::ApplyYV12(BYTE* buf, int pitch, int textcolor, int halocolor, int pitchUV,BYTE* bufU,BYTE* bufV) {
  if (dirty) GetAlphaRect();
  int Ytext = ((textcolor>>16)&255), Utext = ((textcolor>>8)&255), Vtext = (textcolor&255);
  int Yhalo = ((halocolor>>16)&255), Uhalo = ((halocolor>>8)&255), Vhalo = (halocolor&255);
  int w2=w*2;
  char* alpha = alpha_bits;
  for (int y=0; y<(h>>1); y++) {
    for (int x=0; x<w; x+=2) {
      if (*(__int32*)&alpha[x*2] || *(__int32*)&alpha[x*2+w2]) {
        buf[x+0] = (buf[x+0] * (64-alpha[x*2+0]-alpha[x*2+1]) + Ytext * alpha[x*2+0] + Yhalo * alpha[x*2+1]) >> 6;
        buf[x+1] = (buf[x+1] * (64-alpha[x*2+2]-alpha[x*2+3]) + Ytext * alpha[x*2+2] + Yhalo * alpha[x*2+3]) >> 6;
        buf[x+0+pitch] = (buf[x+0+pitch] * (64-alpha[x*2+0+((w2))]-alpha[x*2+1+(w2)]) + Ytext * alpha[x*2+0+(w2)] + Yhalo * alpha[x*2+1+(w2)]) >> 6;
        buf[x+1+pitch] = (buf[x+1+pitch] * (64-alpha[x*2+2+(w2)]-alpha[x*2+3+(w2)]) + Ytext * alpha[x*2+2+(w2)] + Yhalo * alpha[x*2+3+(w2)]) >> 6;

        int auv1 = (alpha[x*2]+alpha[x*2+2]+alpha[x*2+(w2)]+alpha[x*2+2+(w2)])>>1;
        int auv2 = (alpha[x*2+1]+alpha[x*2+3]+alpha[x*2+1+(w2)]+alpha[x*2+3+(w2)])>>1;

        bufU[x>>1] = (bufU[x>>1] * (128-auv1-auv2) + Utext * auv1 + Uhalo * auv2) >> 7;
        bufV[x>>1] = (bufV[x>>1] * (128-auv1-auv2) + Vtext * auv1 + Vhalo * auv2) >> 7;
      }
    }
    buf += pitch*2;
    bufU += pitchUV;
    bufV += pitchUV;
    alpha += w*4;
  }
}


void Antialiaser::ApplyYUY2(BYTE* buf, int pitch, int textcolor, int halocolor) {
  if (dirty) GetAlphaRect();
  int Ytext = ((textcolor>>16)&255), Utext = ((textcolor>>8)&255), Vtext = (textcolor&255);
  int Yhalo = ((halocolor>>16)&255), Uhalo = ((halocolor>>8)&255), Vhalo = (halocolor&255);
  char* alpha = alpha_bits;
  for (int y=h; y>0; --y) {
    for (int x=0; x<w; x+=2) {
      if (*(__int32*)&alpha[x*2]) {
        buf[x*2+0] = (buf[x*2+0] * (64-alpha[x*2+0]-alpha[x*2+1]) + Ytext * alpha[x*2+0] + Yhalo * alpha[x*2+1]) >> 6;
        buf[x*2+2] = (buf[x*2+2] * (64-alpha[x*2+2]-alpha[x*2+3]) + Ytext * alpha[x*2+2] + Yhalo * alpha[x*2+3]) >> 6;
        int auv1 = alpha[x*2]+alpha[x*2+2];
        int auv2 = alpha[x*2+1]+alpha[x*2+3];
        buf[x*2+1] = (buf[x*2+1] * (128-auv1-auv2) + Utext * auv1 + Uhalo * auv2) >> 7;
        buf[x*2+3] = (buf[x*2+3] * (128-auv1-auv2) + Vtext * auv1 + Vhalo * auv2) >> 7;
      }
    }
    buf += pitch;
    alpha += w*2;
  }
}


void Antialiaser::ApplyRGB24(BYTE* buf, int pitch, int textcolor, int halocolor) {
  if (dirty) GetAlphaRect();
  int Rtext = ((textcolor>>16)&255), Gtext = ((textcolor>>8)&255), Btext = (textcolor&255);
  int Rhalo = ((halocolor>>16)&255), Ghalo = ((halocolor>>8)&255), Bhalo = (halocolor&255);
  char* alpha = alpha_bits + (h-1)*w*2;
  for (int y=h; y>0; --y) {
    for (int x=0; x<w; ++x) {
      int textalpha = alpha[x*2+0];
      int haloalpha = alpha[x*2+1];
      if (textalpha | haloalpha) {
        buf[x*3+0] = (buf[x*3+0] * (64-textalpha-haloalpha) + Btext * textalpha + Bhalo * haloalpha) >> 6;
        buf[x*3+1] = (buf[x*3+1] * (64-textalpha-haloalpha) + Gtext * textalpha + Ghalo * haloalpha) >> 6;
        buf[x*3+2] = (buf[x*3+2] * (64-textalpha-haloalpha) + Rtext * textalpha + Rhalo * haloalpha) >> 6;
      }
    }
    buf += pitch;
    alpha -= w*2;
  }
}


void Antialiaser::ApplyRGB32(BYTE* buf, int pitch, int textcolor, int halocolor) {
  if (dirty) GetAlphaRect();
  int Rtext = ((textcolor>>16)&255), Gtext = ((textcolor>>8)&255), Btext = (textcolor&255);
  int Rhalo = ((halocolor>>16)&255), Ghalo = ((halocolor>>8)&255), Bhalo = (halocolor&255);
  char* alpha = alpha_bits + (h-1)*w*2;
  for (int y=h; y>0; --y) {
    for (int x=0; x<w; ++x) {
      int textalpha = alpha[x*2+0];
      int haloalpha = alpha[x*2+1];
      if (textalpha | haloalpha) {
        buf[x*4+0] = (buf[x*4+0] * (64-textalpha-haloalpha) + Btext * textalpha + Bhalo * haloalpha) >> 6;
        buf[x*4+1] = (buf[x*4+1] * (64-textalpha-haloalpha) + Gtext * textalpha + Ghalo * haloalpha) >> 6;
        buf[x*4+2] = (buf[x*4+2] * (64-textalpha-haloalpha) + Rtext * textalpha + Rhalo * haloalpha) >> 6;
      }
    }
    buf += pitch;
    alpha -= w*2;
  }
}


void Antialiaser::GetAlphaRect() {

  dirty = false;

  static BYTE bitcnt[256],    // bit count
              bitexl[256],    // expand to left bit
              bitexr[256];    // expand to right bit
  static bool fInited = false;

  if (!fInited) {
    int i;

    for(i=0; i<256; i++) {
      BYTE b=0, l=0, r=0;

      if (i&  1) { b=1; l|=0x01; r|=0xFF; }
      if (i&  2) { ++b; l|=0x03; r|=0xFE; }
      if (i&  4) { ++b; l|=0x07; r|=0xFC; }
      if (i&  8) { ++b; l|=0x0F; r|=0xF8; }
      if (i& 16) { ++b; l|=0x1F; r|=0xF0; }
      if (i& 32) { ++b; l|=0x3F; r|=0xE0; }
      if (i& 64) { ++b; l|=0x7F; r|=0xC0; }
      if (i&128) { ++b; l|=0xFF; r|=0x80; }

      bitcnt[i] = b;
      bitexl[i] = l;
      bitexr[i] = r;
    }

    fInited = true;
  }

  int srcpitch = (w+4+3) & -4;

  char* dst = alpha_bits;
  for (int y=0; y<h; ++y) {
    BYTE* src = (BYTE*)lpAntialiasBits + ((h-y-1)*8 + 20) * srcpitch + 2;
    int wt = w;
    do {
      int alpha1, alpha2;
      int i;
      BYTE bmasks[8], tmasks[8];

      alpha1 = alpha2 = 0;

/*      BYTE tmp = 0;
      for (i=0; i<8; i++) {
        tmp |= src[srcpitch*i];
        tmp |= src[srcpitch*i-1];
        tmp |= src[srcpitch*i+1];
        tmp |= src[srcpitch*(-8+i)];
        tmp |= src[srcpitch*(-8+i)-1];
        tmp |= src[srcpitch*(-8+i)+1];
        tmp |= src[srcpitch*(8+i)];
        tmp |= src[srcpitch*(8+i)-1];
        tmp |= src[srcpitch*(8+i)+1];
      }
*/
      DWORD tmp;
      __asm {           // test if the whole area isn't just plain black
        mov edx, srcpitch
        mov esi, src
        mov ecx, edx
        dec esi
        shl ecx, 3
        sub esi, ecx  ; src - 8*pitch - 1

        mov eax, [esi]  ; repeat 24 times
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]
        add esi, edx
        or eax, [esi]

        and eax, 0x00ffffff
        mov tmp, eax
      }

      if (tmp != 0) {     // quick exit in a common case
        for (i=0; i<8; i++)
          alpha1 += bitcnt[src[srcpitch*i]];

        BYTE cenmask = 0, mask1, mask2;

        for(i=0; i<=8; i++) {
          cenmask |= (BYTE)(((long)-src[srcpitch*i  ])>>31);
          cenmask |= bitexl[src[srcpitch*i-1]];
          cenmask |= bitexr[src[srcpitch*i+1]];
        }

        mask1 = mask2 = cenmask;

        for(i=0; i<8; i++) {
          mask1 |= (BYTE)(((long)-src[srcpitch*(-i)])>>31);
          mask1 |= bitexl[src[srcpitch*(-8+i)-1]];
          mask1 |= bitexr[src[srcpitch*(-8+1)+1]];
          mask2 |= (BYTE)(((long)-src[srcpitch*(8+i)])>>31);
          mask2 |= bitexl[src[srcpitch*(8+i)-1]];
          mask2 |= bitexr[src[srcpitch*(8+i)+1]];

          tmasks[i] = mask1;
          bmasks[i] = mask2;
        }

        for(i=0; i<8; i++) {
          alpha2 += bitcnt[cenmask | tmasks[7-i] | bmasks[i]];
        }
      }

      dst[0] = alpha1;
      dst[1] = alpha2-alpha1;
      dst += 2;
      ++src;
    } while(--wt);
  }
}

